#
# Copyright  1990-2006 Sun Microsystems, Inc. All Rights Reserved.
# DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version
# 2 only, as published by the Free Software Foundation. 
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License version 2 for more details (a copy is
# included at /legal/license.txt). 
# 
# You should have received a copy of the GNU General Public License
# version 2 along with this work; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
# 02110-1301 USA 
# 
# Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa
# Clara, CA 95054 or visit www.sun.com if you need additional
# information or have any questions. 
#
# @(#)defs_jit_risc.mk	1.27 06/10/13
#

#
# defs for the shared risc-based jit code
#


#
# All the risc jit specific build flags we need to keep track of in
# case they are toggled.
#
CVM_JIT_COPY_CCMCODE_TO_CODECACHE ?= false
CVM_JIT_PMI ?= false

CVM_FLAGS += \
	CVM_JIT_COPY_CCMCODE_TO_CODECACHE \
	CVM_JIT_PMI \

CVM_JIT_COPY_CCMCODE_TO_CODECACHE_CLEANUP_ACTION = \
	$(CVM_DEFAULT_CLEANUP_ACTION)
CVM_JIT_PMI_CLEANUP_ACTION = \
	$(CVM_JIT_CLEANUP_ACTION)

#
# If your port wants all of the ccm assembler code to be copied to the
# code cache, the set CVM_JIT_COPY_CCMCODE_TO_CODECACHE to true.
#
ifeq ($(CVM_JIT_COPY_CCMCODE_TO_CODECACHE), true)
CVM_DEFINES	+= -DCVM_JIT_COPY_CCMCODE_TO_CODECACHE
endif

ifeq ($(CVM_JIT_PMI), true)
CVM_DEFINES   += -DCVM_JIT_PATCHED_METHOD_INVOCATIONS
endif

ifdef CVM_JIT_PATCHED_METHOD_INVOCATIONS
$(error CVM_JIT_PATCHED_METHOD_INVOCATIONS is no longer supported.\
	Use CVM_JIT_PMI instead.)
endif

# irparser files generated by jcs
CVM_TARGETOBJS_SPACE += \
	jitcodegen.o \
	jitcodegentable.o \

CVM_TARGETOBJS_SPEED +=	\
	jitregman.o \
        ccmintrinsics_risc.o \

CVM_TARGETOBJS_SPACE += \
	jitfixup.o \
	jitstackman.o \
	jitopcodes.o \
	jitemitter_cpu.o \
	jitinit_cpu.o \
	jit_risc.o \
	jitemitter.o \

JIT_CPU_O		?= jit_cpu.o
CCMCODECACHECOPY_CPU_O	?= ccmcodecachecopy_cpu.o
CVM_TARGETOBJS_OTHER += \
	$(JIT_CPU_O) \
	$(CCMCODECACHECOPY_CPU_O)

CVM_SRCDIRS   += \
	$(CVM_TOP)/src/$(TARGET_CPU_FAMILY)/javavm/runtime/jit \
	$(CVM_TOP)/src/portlibs/jit/risc

CVM_INCLUDE_DIRS  += \
	$(CVM_TOP)/src/portlibs/jit/risc

#
# JCS input files. Besides needing shared and cpu specific versions, they
# also need to be split up further so they various parts can be combined
# in the proper order.
#

CVM_JCS_SHARED_INCLUDES_FILE = \
    $(CVM_TOP)/src/portlibs/jit/risc/jitgrammarincludes.jcs

CVM_JCS_SHARED_DEFS_FILE     = \
    $(CVM_TOP)/src/portlibs/jit/risc/jitgrammardefs.jcs

CVM_JCS_SHARED_RULES_FILE    = \
    $(CVM_TOP)/src/portlibs/jit/risc/jitgrammarrules.jcs

ifeq ($(CVM_JIT_USE_FP_HARDWARE), true)
CVM_JCS_SHARED_RULES_FILE    += \
    $(CVM_TOP)/src/portlibs/jit/risc/jitfloatgrammarrules.jcs
endif

# The following should be defined in CPU specific makefiles:
#CVM_JCS_CPU_INCLUDES_FILE =
#CVM_JCS_CPU_DEFS_FILE     =
#CVM_JCS_CPU_RULES_FILE    = 

CVM_JCS_INPUT_FILES = \
	$(CVM_JCS_SHARED_INCLUDES_FILE) \
	$(CVM_JCS_CPU_INCLUDES_FILE) \
	$(CVM_JCS_SHARED_DEFS_FILE) \
	$(CVM_JCS_CPU_DEFS_FILE) \
	$(CVM_JCS_SHARED_RULES_FILE) \
	$(CVM_JCS_CPU_RULES_FILE) \
