/*
 * @(#)TestMemmove.c	1.6 06/10/10
 *
 * Copyright  1990-2006 Sun Microsystems, Inc. All Rights Reserved.  
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER  
 *   
 * This program is free software; you can redistribute it and/or  
 * modify it under the terms of the GNU General Public License version  
 * 2 only, as published by the Free Software Foundation.   
 *   
 * This program is distributed in the hope that it will be useful, but  
 * WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  
 * General Public License version 2 for more details (a copy is  
 * included at /legal/license.txt).   
 *   
 * You should have received a copy of the GNU General Public License  
 * version 2 along with this work; if not, write to the Free Software  
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  
 * 02110-1301 USA   
 *   
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa  
 * Clara, CA 95054 or visit www.sun.com if you need additional  
 * information or have any questions. 
 *
 */
/*
 * @(#)TestMemmove.c	1.2 03/01/29
 * Compliance test for Java memmove functions used by CVM's porting layer.
 */
#include "javavm/include/porting/memory.h"

#define ARRAY_SIZE(x)   (sizeof(x)/sizeof(x[0]))

static void
initArray(CVMUint8 *array, int size, int start)
{
    int i;
    for (i = 0; i < size; i++) {
        array[i] = start + i + 1;
    }
}

static int testCount = 0;
static int testPassed = 0;

static CVMBool
verifyBytes(const CVMUint8 *testname,
            const CVMUint8 *array, const CVMUint8 *expected, int size)
{
    int i;
    testCount++;
    for (i = 0; i < size; i++) {
        if (array[i] != expected[i]) {
            int j;
            CVMconsolePrintf("FAILED: %s\n", testname);
            CVMconsolePrintf("       : Error starting at index %d\n", i);
            CVMconsolePrintf("       : expecting %d, actual %d\n",
                             expected[i], array[i]);
            CVMconsolePrintf("       : Expecting:\n");
            for (j = 0; j < size; j++) {
                if ((j % 8) == 0) {
                    CVMconsolePrintf("       :");
                }
                CVMconsolePrintf(" %3d,", expected[j]);
                if ((j % 8) == 7) {
                    CVMconsolePrintf("\n");
                }
            }
            if ((j % 8) != 0) {
                CVMconsolePrintf("\n");
            }
            CVMconsolePrintf("       : Actual:\n");
            for (j = 0; j < size; j++) {
                if ((j % 8) == 0) {
                    CVMconsolePrintf("       :");
                }
                CVMconsolePrintf(" %3d,", array[j]);
                if ((j % 8) == 7) {
                    CVMconsolePrintf("\n");
                }
            }
            if ((j % 8) != 0) {
                CVMconsolePrintf("\n");
            }
            return CVM_FALSE;
        }
    }
    testPassed++;
    CVMconsolePrintf("PASSED: %s\n", testname);
    return CVM_TRUE;
}

static void
testByteMemmove(int salign, int dalign, int length,
                const CVMUint8 *expected, int expected_size,
                CVMBool isByte)
{
    CVMUint8 src[150];
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *srcAlignedStart;
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for src and dest arrays: */
    srcAlignedStart = (CVMUint8 *)(((CVMUint32)src + 7) & ~7);
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(srcAlignedStart, 100, 100);
        initArray(destAlignedStart, 100, 0);
        s = srcAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isByte) {
            CVMmemmoveByte(d, s, length);
            sprintf(testname, "byte: salign %d dalign %d length %d",
                    salign, dalign, length);
        } else {
            CVMmemmoveBoolean(d, s, length);
            sprintf(testname, "boolean: salign %d dalign %d length %d",
                    salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testCharMemmove(int salign, int dalign, int length,
                const CVMUint8 *expected, int expected_size,
                CVMBool isChar)
{
    CVMUint8 src[150];
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *srcAlignedStart;
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for src and dest arrays: */
    srcAlignedStart = (CVMUint8 *)(((CVMUint32)src + 7) & ~7);
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(srcAlignedStart, 100, 100);
        initArray(destAlignedStart, 100, 0);
        s = srcAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isChar) {
            CVMmemmoveChar(d, s, length);
            sprintf(testname, "char: salign %d dalign %d length %d",
                    salign, dalign, length);
        } else {
            CVMmemmoveShort(d, s, length);
            sprintf(testname, "short: salign %d dalign %d length %d",
                    salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testIntMemmove(int salign, int dalign, int length,
               const CVMUint8 *expected, int expected_size,
               CVMBool isInt)
{
    CVMUint8 src[150];
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *srcAlignedStart;
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for src and dest arrays: */
    srcAlignedStart = (CVMUint8 *)(((CVMUint32)src + 7) & ~7);
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(srcAlignedStart, 100, 100);
        initArray(destAlignedStart, 100, 0);
        s = srcAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isInt) {
            CVMmemmoveInt(d, s, length);
            sprintf(testname, "int: salign %d dalign %d length %d",
                    salign, dalign, length);
        } else {
            CVMmemmoveFloat(d, s, length);
            sprintf(testname, "float: salign %d dalign %d length %d",
                    salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testLongMemmove(int salign, int dalign, int length,
                const CVMUint8 *expected, int expected_size,
                CVMBool isDouble)
{
    CVMUint8 src[150];
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *srcAlignedStart;
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for src and dest arrays: */
    srcAlignedStart = (CVMUint8 *)(((CVMUint32)src + 7) & ~7);
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(srcAlignedStart, 100, 100);
        initArray(destAlignedStart, 100, 0);
        s = srcAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isDouble) {
            CVMmemmoveLong(d, s, length);
            sprintf(testname, "long: salign %d dalign %d length %d",
                    salign, dalign, length);
        } else {
            CVMmemmoveDouble(d, s, length);
            sprintf(testname, "double: salign %d dalign %d length %d",
                    salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testOByteMemmove(int salign, int dalign, int length,
                 const CVMUint8 *expected, int expected_size,
                 CVMBool isByte)
{
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for the dest array: */
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(destAlignedStart, 100, 0);
        s = destAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isByte) {
            CVMmemmoveByte(d, s, length);
            sprintf(testname,
                "overlapping byte: salign %d dalign %d length %d",
                salign, dalign, length);
        } else {
            CVMmemmoveBoolean(d, s, length);
            sprintf(testname,
                "overlapping boolean: salign %d dalign %d length %d",
                salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testOCharMemmove(int salign, int dalign, int length,
                 const CVMUint8 *expected, int expected_size,
                 CVMBool isChar)
{
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for the dest array: */
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(destAlignedStart, 100, 0);
        s = destAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isChar) {
            CVMmemmoveChar(d, s, length);
            sprintf(testname,
                "overlapping char: salign %d dalign %d length %d",
                salign, dalign, length);
        } else {
            CVMmemmoveShort(d, s, length);
            sprintf(testname,
                "overlapping short: salign %d dalign %d length %d",
                salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testOIntMemmove(int salign, int dalign, int length,
                const CVMUint8 *expected, int expected_size,
                CVMBool isInt)
{
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for the dest array: */
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(destAlignedStart, 100, 0);
        s = destAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isInt) {
            CVMmemmoveInt(d, s, length);
            sprintf(testname,
                "overlapping int: salign %d dalign %d length %d",
                salign, dalign, length);
        } else {
            CVMmemmoveFloat(d, s, length);
            sprintf(testname,
                "overlapping float: salign %d dalign %d length %d",
                salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testOLongMemmove(int salign, int dalign, int length,
                 const CVMUint8 *expected, int expected_size,
                 CVMBool isLong)
{
    CVMUint8 dest[150];
    CVMUint8 testname[150];
    CVMUint8 *destAlignedStart;

    /* Find the 64-bit aligned starting point for the dest array: */
    destAlignedStart = (CVMUint8 *)(((CVMUint32)dest + 7) & ~7);

    {
        CVMUint8 *s, *d;
        initArray(destAlignedStart, 100, 0);
        s = destAlignedStart + salign + 8;
        d = destAlignedStart + dalign + 8;
        if (isLong) {
            CVMmemmoveLong(d, s, length);
            sprintf(testname,
                "overlapping long: salign %d dalign %d length %d",
                salign, dalign, length);
        } else {
            CVMmemmoveDouble(d, s, length);
            sprintf(testname,
                "overlapping double: salign %d dalign %d length %d",
                salign, dalign, length);
        }
        verifyBytes(testname,
            destAlignedStart, expected, expected_size);
    }
}

static void
testByteBoolean(CVMBool isByte)
{
    int salign, dalign;

    /* Byte: Align s0 d0: ================================================= */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            117, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s0 d1: ================================================= */
    salign = 0;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 114, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 114, 115,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 114, 115,
            116, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 114, 115,
            116, 117, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 114, 115,
            116, 117, 118, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 109, 110, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s0 d2: ================================================= */
    salign = 0;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 114,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 114,
            115, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 114,
            115, 116, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 114,
            115, 116, 117, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 114,
            115, 116, 117, 118, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 120, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s0 d3: ================================================= */
    salign = 0;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            114, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            114, 115, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            114, 115, 116, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            114, 115, 116, 117, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            114, 115, 116, 117, 118, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            114, 115, 116, 117, 118, 119, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 109, 110, 111, 112, 113,
            114, 115, 116, 117, 118, 119, 120, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s0 d4: ================================================= */
    salign = 0;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 117, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 117, 118, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 117, 118, 119, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 117, 118, 119, 120,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }



    /* Byte: Align s1 d0: ================================================= */
    salign = 1;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 115, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 115, 116, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 115, 116, 117,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 115, 116, 117,
            118, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 115, 116, 117,
            118, 119, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            110, 111, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s1 d1: ================================================= */
    salign = 1;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 115, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 115, 116,
            117, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s1 d2: ================================================= */
    salign = 1;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 115,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 115,
            116, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 115,
            116, 117, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 115,
            116, 117, 118, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 110, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 121, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s1 d3: ================================================= */
    salign = 1;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            115, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            115, 116, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            115, 116, 117, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            115, 116, 117, 118, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 120, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 110, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 120, 121, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s1 d4: ================================================= */
    salign = 1;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 115, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 115, 116, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 115, 116, 117, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 115, 116, 117, 118, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 115, 116, 117, 118, 119, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 115, 116, 117, 118, 119, 120, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 110, 111, 112, 113,
            114, 115, 116, 117, 118, 119, 120, 121,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s2 d0: ================================================= */
    salign = 2;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 117, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 117, 118,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 117, 118,
            119, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 117, 118,
            119, 120, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 122, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s2 d1: ================================================= */
    salign = 2;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  111, 112, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  111, 112, 113, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 116, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 116, 117,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 116, 117,
            118, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 116, 117,
            118, 119, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 111, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 122, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s2 d2: ================================================= */
    salign = 2;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 116,
            117, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s2 d3: ================================================= */
    salign = 2;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            116, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            116, 117, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            116, 117, 118, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 121, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 111, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 121, 122, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s2 d4: ================================================= */
    salign = 2;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 116, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 116, 117, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 116, 117, 118, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 120, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 120, 121, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            115, 116, 117, 118, 119, 120, 121, 122,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s3 d0: ================================================= */
    salign = 3;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 117, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 117, 118, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 117, 118, 119,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 117, 118, 119,
            120, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 117, 118, 119,
            120, 121, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 117, 118, 119,
            120, 121, 122, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            112, 113, 114, 115, 116, 117, 118, 119,
            120, 121, 122, 123, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s3 d1: ================================================= */
    salign = 3;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  112, 113, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  112, 113, 114, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 117, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 117, 118,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 117, 118,
            119, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 117, 118,
            119, 120, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 122, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 112, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 122, 123, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s3 d2: ================================================= */
    salign = 3;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 117,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 117,
            118, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 117,
            118, 119, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 122, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 112, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 122, 123, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s3 d3: ================================================= */
    salign = 3;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            117, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            117, 118, 119, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 123, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s3 d4: ================================================= */
    salign = 3;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 117, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 117, 118, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 117, 118, 119, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 121, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 121, 122, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 112, 113, 114, 115,
            116, 117, 118, 119, 120, 121, 122, 123,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s4 d0: ================================================= */
    salign = 4;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 122, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 122, 123, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 122, 123, 124, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s4 d1: ================================================= */
    salign = 4;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  113, 114, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  113, 114, 115, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 118, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 118, 119,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 118, 119,
            120, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 118, 119,
            120, 121, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 118, 119,
            120, 121, 122, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 118, 119,
            120, 121, 122, 123, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 113, 114, 115, 116, 117, 118, 119,
            120, 121, 122, 123, 124, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s4 d2: ================================================= */
    salign = 4;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 118,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 118,
            119, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 118,
            119, 120, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 122, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 122, 123, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 117, 118,
            119, 120, 121, 122, 123, 124, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s4 d3: ================================================= */
    salign = 4;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            118, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            118, 119, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            118, 119, 120, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 122, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 122, 123, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 113, 114, 115, 116, 117,
            118, 119, 120, 121, 122, 123, 124, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    /* Byte: Align s4 d4: ================================================= */
    salign = 4;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 123, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                        isByte);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 123, 124,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                        isByte);
    }

}

static void
testCharShort(CVMBool isChar)
{
    int salign, dalign;

    /* Byte: Align s0 d0: ================================================= */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s0 d2: ================================================= */
    salign = 0;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 109, 110, 111, 112, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 109, 110, 111, 112, 113, 114,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 109, 110, 111, 112, 113, 114,
            115, 116, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s0 d4: ================================================= */
    salign = 0;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 109, 110, 111, 112,
            113, 114, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s0 d6: ================================================= */
    salign = 0;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 109, 110,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 109, 110,
            111, 112, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 109, 110,
            111, 112, 113, 114, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 109, 110,
            111, 112, 113, 114, 115, 116, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s0 d8: ================================================= */
    salign = 0;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            109, 110, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            109, 110, 111, 112, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 15, 16,
            109, 110, 111, 112, 113, 114, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            109, 110,111, 112, 113, 114, 115, 116,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s2 d0: ================================================= */
    salign = 2;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            111, 112, 113, 114, 115, 116, 117, 118,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s2 d2: ================================================= */
    salign = 2;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 111, 112, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 111, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 111, 112, 113, 114, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s2 d4: ================================================= */
    salign = 2;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 111, 112, 113, 114,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 111, 112, 113, 114, 
            115, 116, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 111, 112, 113, 114,
            115, 116, 117, 118, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s2 d6: ================================================= */
    salign = 2;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 111, 112,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 111, 112, 
            113, 114, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 111, 112, 
            113, 114, 115, 116, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 111, 112,
            113, 114, 115, 116, 117, 118, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s2 d8: ================================================= */
    salign = 2;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            111, 112, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            111, 112, 113, 114, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 15, 16,
            111, 112, 113, 114, 115, 116, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            111, 112, 113, 114, 115, 116, 117, 118,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s4 d0: ================================================= */
    salign = 4;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s4 d2: ================================================= */
    salign = 4;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 113, 114, 115, 116, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 113, 114, 115, 116, 117, 118, 
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 113, 114, 115, 116, 117, 118, 
            119, 120, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s4 d4: ================================================= */
    salign = 4;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 113, 114, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s4 d6: ================================================= */
    salign = 4;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 113, 114,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 113, 114, 
            115, 116, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 113, 114,
            115, 116, 117, 118, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 113, 114, 
            115, 116, 117, 118, 119, 120, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s4 d8: ================================================= */
    salign = 4;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            113, 114, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            113, 114, 115, 116, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 15, 16,
            113, 114, 115, 116, 117, 118, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            113, 114, 115, 116, 117, 118, 119, 120,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s6 d0: ================================================= */
    salign = 6;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            115, 116, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            115, 116, 117, 118, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            115, 116, 117, 118, 119, 120, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            115, 116, 117, 118, 119, 120, 121, 122,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s6 d2: ================================================= */
    salign = 6;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 115, 116, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 115, 116, 117, 118, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 115, 116, 117, 118, 119, 120,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 115, 116, 117, 118, 119, 120,
            121, 122, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s6 d4: ================================================= */
    salign = 6;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 115, 116, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 115, 116, 117, 118,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 115, 116, 117, 118,
            119, 120, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 115, 116, 117, 118,
            119, 120, 121, 122, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s6 d6: ================================================= */
    salign = 6;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 115, 116,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 115, 116,
            117, 118, 119, 120, 121, 122, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s6 d8: ================================================= */
    salign = 6;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            115, 116, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            115, 116, 117, 118, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 15, 16,
            115, 116, 117, 118, 119, 120, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            115, 116, 117, 118, 119, 120, 121, 122,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s8 d0: ================================================= */
    salign = 8;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 119, 120, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 119, 120, 121, 122, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 119, 120, 121, 122, 123, 124, 
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s8 d2: ================================================= */
    salign = 8;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 117, 118, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 117, 118, 119, 120, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 117, 118, 119, 120, 121, 122,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 117, 118, 119, 120, 121, 122,
            123, 124, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s8 d4: ================================================= */
    salign = 8;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 117, 118, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 117, 118, 119, 120,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 117, 118, 119, 120,
            121, 122, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 117, 118, 119, 120,
            121, 122, 123, 124, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s8 d6: ================================================= */
    salign = 8;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 117, 118,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 117, 118,
            119, 120, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 117, 118,
            119, 120, 121, 122, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 117, 118,
            119, 120, 121, 122, 123, 124, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    /* Byte: Align s8 d8: ================================================= */
    salign = 8;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            117, 118, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10,  11, 12, 13, 14, 15, 16,
            117, 118, 119, 120, 121, 122, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                        isChar);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            117, 118, 119, 120, 121, 122, 123, 124,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isChar);
    }

}

static void
testIntFloat(CVMBool isInt)
{
    int salign, dalign;

    /* Byte: Align s0 d0: ================================================= */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s0 d4: ================================================= */
    salign = 0;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s0 d8: ================================================= */
    salign = 0;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            109, 110, 111, 112, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            109, 110, 111, 112, 113, 114, 115, 116,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s4 d0: ================================================= */
    salign = 4;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120, 
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s4 d4: ================================================= */
    salign = 4;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s4 d8: ================================================= */
    salign = 4;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            113, 114, 115, 116, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            113, 114, 115, 116, 117, 118, 119, 120,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s8 d0: ================================================= */
    salign = 8;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 119, 120, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 119, 120, 121, 122, 123, 124,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s8 d4: ================================================= */
    salign = 8;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 117, 118, 119, 120,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 117, 118, 119, 120,
            121, 122, 123, 124, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    /* Byte: Align s8 d8: ================================================= */
    salign = 8;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                       isInt);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            117, 118, 119, 120, 121, 122, 123, 124,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                       isInt);
    }

}

static void
testLongDouble(CVMBool isLong)
{
    int salign, dalign;

    /* Byte: Align s0 d0: ================================================= */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            109, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 123, 124,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s0 d4: ================================================= */
    salign = 0;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 109, 110, 111, 112,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 122, 123, 124, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s0 d8: ================================================= */
    salign = 0;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            109, 110, 111, 112, 113, 114, 115, 116,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            109, 110, 111, 112, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 123, 124,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s0 d12: ================================================ */
    salign = 0;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 109, 110, 111, 112,
            113, 114, 115, 116, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 109, 110, 111, 112,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 122, 123, 124, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s4 d0: ================================================= */
    salign = 4;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 122, 123, 124, 125, 126, 127, 128, 
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s4 d4: ================================================= */
    salign = 4;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 123, 124,
            125, 126, 127, 128, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s4 d8: ================================================= */
    salign = 4;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            113, 114, 115, 116, 117, 118, 119, 120,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            113, 114, 115, 116, 117, 118, 119, 120,
            121, 122, 123, 124, 125, 126, 127, 128,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s4 d12: ================================================ */
    salign = 4;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 113, 114, 115, 116,
            117, 118, 119, 120, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 113, 114, 115, 116,
            117, 118, 119, 120, 121, 122, 123, 124,
            125, 126, 127, 128, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s8 d0: ================================================= */
    salign = 8;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 119, 120, 121, 122, 123, 124,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            117, 118, 119, 120, 121, 122, 123, 124,
            125, 126, 127, 128, 129, 130, 131, 132,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s8 d4: ================================================= */
    salign = 8;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 117, 118, 119, 120,
            121, 122, 123, 124, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 117, 118, 119, 120,
            121, 122, 123, 124, 125, 126, 127, 128,
            129, 130, 131, 132, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s8 d8: ================================================= */
    salign = 8;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            117, 118, 119, 120, 121, 122, 123, 124,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            117, 118, 119, 120, 121, 122, 123, 124,
            125, 126, 127, 128, 129, 130, 131, 132,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s8 d12: ================================================ */
    salign = 8;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 117, 118, 119, 120,
            121, 122, 123, 124, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 117, 118, 119, 120,
            121, 122, 123, 124, 125, 126, 127, 128,
            129, 130, 131, 132, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s12 d0: ================================================= */
    salign = 12;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            121, 122, 123, 124, 125, 126, 127, 128,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            121, 122, 123, 124, 125, 126, 127, 128,
            129, 130, 131, 132, 133, 134, 135, 136,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s12 d4: ================================================= */
    salign = 12;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 121, 122, 123, 124,
            125, 126, 127, 128, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 121, 122, 123, 124,
            125, 126, 127, 128, 129, 130, 131, 132,
            133, 134, 135, 136, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s12 d8: ================================================= */
    salign = 12;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            121, 122, 123, 124, 125, 126, 127, 128,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            121, 122, 123, 124, 125, 126, 127, 128,
            129, 130, 131, 132, 133, 134, 135, 136,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    /* Byte: Align s12 d12: ================================================ */
    salign = 12;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 121, 122, 123, 124,
            125, 126, 127, 128, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                        isLong);
    }

    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 121, 122, 123, 124,
            125, 126, 127, 128, 129, 130, 131, 132,
            133, 134, 135, 136, 37, 38, 39, 40,
        };
        testLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                        isLong);
    }

}

/* ///////////////////////////////////////////////////////////////////////// */

static void
testOverlappingByteBoolean(CVMBool isByte)
{
    int salign, dalign;

    /* Byte: Align s0 d0: ================================================= */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d1: ================================================= */
    salign = 0;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 14, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 14, 15,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 14, 15,
            16, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 14, 15,
            16, 17, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 14, 15,
            16, 17, 18, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  9,  10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d2: ================================================= */
    salign = 0;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 14,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 14,
            15, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 14,
            15, 16, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 18, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d3: ================================================= */
    salign = 0;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            14, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            14, 15, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            14, 15, 16, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 18, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d4: ================================================= */
    salign = 0;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 14, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 14, 15, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 14, 15, 16, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 18, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 20,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d5: ================================================= */
    salign = 0;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 14, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 14, 15, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 14, 15, 16, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 14, 15, 16, 17, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 14, 15, 16, 17, 18, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 9,  10, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            20, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d6: ================================================= */
    salign = 0;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 14, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 14, 15, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 14, 15, 16, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 14, 15, 16, 17, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 14, 15, 16, 17, 18,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 9,  10,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d7: ================================================= */
    salign = 0;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 14, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 14, 15, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 14, 15, 16, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 14, 15, 16, 17,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 9,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d8: ================================================= */
    salign = 0;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 14, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 14, 15, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 14, 15, 16,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d9: ================================================= */
    salign = 0;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 14, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 14, 15,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 14, 15,
            16, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 14, 15,
            16, 17, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 14, 15,
            16, 17, 18, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 9,  10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d10: ================================================= */
    salign = 0;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 14,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 14,
            15, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 14,
            15, 16, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 18, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 9,  10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d11: ================================================= */
    salign = 0;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            14, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            14, 15, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            14, 15, 16, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 18, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 9,  10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s0 d12: ================================================= */
    salign = 0;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 14, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 14, 15, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 14, 15, 16, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 18, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 9,  10, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 20,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d0: ================================================= */
    salign = 1;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 15, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 15, 16, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 15, 16, 17,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 21, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d1: ================================================= */
    salign = 1;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d2: ================================================= */
    salign = 1;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 15,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 15,
            16, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 15,
            16, 17, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 21, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d3: ================================================= */
    salign = 1;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            15, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            15, 16, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            15, 16, 17, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d4: ================================================= */
    salign = 1;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 15, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 15, 16, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 15, 16, 17, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 21,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d5: ================================================= */
    salign = 1;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 15, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 15, 16, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 15, 16, 17, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 15, 16, 17, 18, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 20,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 10, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 20,
            21, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d6: ================================================= */
    salign = 1;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 15, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 15, 16, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 15, 16, 17, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 15, 16, 17, 18, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            20, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 10, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            20, 21, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d7: ================================================= */
    salign = 1;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 15, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 15, 16, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 15, 16, 17, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 15, 16, 17, 18,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 10,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 21, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d8: ================================================= */
    salign = 1;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 15, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 15, 16, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 15, 16, 17,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            10, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 21, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d9: ================================================= */
    salign = 1;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 15, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 15, 16,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 15, 16,
            17, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d10: ================================================= */
    salign = 1;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 15,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 15,
            16, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 15,
            16, 17, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 10, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 21, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d11: ================================================= */
    salign = 1;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            15, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            15, 16, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            15, 16, 17, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 10, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s1 d12: ================================================= */
    salign = 1;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 15, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 15, 16, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 15, 16, 17, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 10, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 21,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }


    /* Byte: Align s2 d0: ================================================= */
    salign = 2;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 16, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 16, 17, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 16, 17, 18,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 21, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 21, 22, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d1: ================================================= */
    salign = 2;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 16, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 16, 17,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 16, 17,
            18, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 16, 17,
            18, 19, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 21, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 21, 22, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d2: ================================================= */
    salign = 2;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d3: ================================================= */
    salign = 2;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            16, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            16, 17, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            16, 17, 18, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 21, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 21, 22, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d4: ================================================= */
    salign = 2;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 16, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 16, 17, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 16, 17, 18, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 22,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d5: ================================================= */
    salign = 2;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 16, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 16, 17, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 16, 17, 18, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 21,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 11, 12, 13,
            14, 15, 16, 17, 18, 19, 20, 21,
            22, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d6: ================================================= */
    salign = 2;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 16, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 16, 17, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 16, 17, 18, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 20,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 20,
            21, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 11, 12,
            13, 14, 15, 16, 17, 18, 19, 20,
            21, 22, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d7: ================================================= */
    salign = 2;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 16, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 16, 17, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 16, 17, 18, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            20, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            20, 21, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 11,
            12, 13, 14, 15, 16, 17, 18, 19,
            20, 21, 22, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d8: ================================================= */
    salign = 2;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 16, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 16, 17, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 16, 17, 18,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 21, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            11, 12, 13, 14, 15, 16, 17, 18,
            19, 20, 21, 22, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d9: ================================================= */
    salign = 2;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 16, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 16, 17,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 16, 17,
            18, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 21, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 11, 12, 13, 14, 15, 16, 17,
            18, 19, 20, 21, 22, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d10: ================================================= */
    salign = 2;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 16,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 16,
            17, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 16,
            17, 18, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d11: ================================================= */
    salign = 2;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            16, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            16, 17, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            16, 17, 18, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 21, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 11, 12, 13, 14, 15,
            16, 17, 18, 19, 20, 21, 22, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s2 d12: ================================================= */
    salign = 2;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 21, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 22, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 23, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 24,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            25, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 26, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 16, 27, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 16, 17, 28, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 16, 17, 18, 29, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 30, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 31, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 32,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1,  2,  3,  4,  5,  6,  7,  8,
            9,  10, 11, 12, 13, 14, 15, 16,
            17, 18, 19, 20, 11, 12, 13, 14,
            15, 16, 17, 18, 19, 20, 21, 22,
            33, 34, 35, 36, 37, 38, 39, 40,
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d0: ================================================= */
    salign = 3;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 17, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 17, 18, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d1: ================================================= */
    salign = 3;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 17, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 17, 18, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d2: ================================================= */
    salign = 3;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 17, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 17, 
            18, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 17, 
            18, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d3: ================================================= */
    salign = 3;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d4: ================================================= */
    salign = 3;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 17, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 17, 18, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 17, 18, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d5: ================================================= */
    salign = 3;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 17, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 17, 18, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d6: ================================================= */
    salign = 3;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 17, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 17, 18, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 17, 18, 19, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 17, 18, 19, 20, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 12, 13, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d7: ================================================= */
    salign = 3;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 17, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 17, 18, 19, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d8: ================================================= */
    salign = 3;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 17, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 17, 18, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            12, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d9: ================================================= */
    salign = 3;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 17, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 17, 18, 
            19, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d10: ================================================= */
    salign = 3;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 17, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 17, 
            18, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 17, 
            18, 19, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 12, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d11: ================================================= */
    salign = 3;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s3 d12: ================================================= */
    salign = 3;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 17, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 17, 18, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 17, 18, 19, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 12, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d0: ================================================= */
    salign = 4;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d1: ================================================= */
    salign = 4;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 18, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 18, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 18, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d2: ================================================= */
    salign = 4;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d3: ================================================= */
    salign = 4;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            18, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            18, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            18, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d4: ================================================= */
    salign = 4;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d5: ================================================= */
    salign = 4;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 18, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 18, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 18, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 13, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d6: ================================================= */
    salign = 4;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d7: ================================================= */
    salign = 4;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 18, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 18, 19, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 18, 19, 20, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 13, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d8: ================================================= */
    salign = 4;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d9: ================================================= */
    salign = 4;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 18, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 18, 19, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 18, 19, 
            20, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 13, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d10: ================================================= */
    salign = 4;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d11: ================================================= */
    salign = 4;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            18, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            18, 19, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            18, 19, 20, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 13, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s4 d12: ================================================= */
    salign = 4;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d0: ================================================= */
    salign = 5;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 19, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 19, 20, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d1: ================================================= */
    salign = 5;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 19, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 19, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d2: ================================================= */
    salign = 5;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d3: ================================================= */
    salign = 5;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d4: ================================================= */
    salign = 5;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d5: ================================================= */
    salign = 5;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d6: ================================================= */
    salign = 5;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 14, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d7: ================================================= */
    salign = 5;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d8: ================================================= */
    salign = 5;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 19, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 19, 20, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            14, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d9: ================================================= */
    salign = 5;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 19, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 19, 20, 
            21, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d10: ================================================= */
    salign = 5;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 19, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 19, 
            20, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 19, 
            20, 21, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 14, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d11: ================================================= */
    salign = 5;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            19, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            19, 20, 21, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s5 d12: ================================================= */
    salign = 5;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 19, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 19, 20, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 19, 20, 21, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 14, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d0: ================================================= */
    salign = 6;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d1: ================================================= */
    salign = 6;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 20, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 20, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 20, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d2: ================================================= */
    salign = 6;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d3: ================================================= */
    salign = 6;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d4: ================================================= */
    salign = 6;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d5: ================================================= */
    salign = 6;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 15, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d6: ================================================= */
    salign = 6;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d7: ================================================= */
    salign = 6;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 15, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d8: ================================================= */
    salign = 6;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d9: ================================================= */
    salign = 6;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 20, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 20, 21, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 20, 21, 
            22, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 15, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d10: ================================================= */
    salign = 6;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d11: ================================================= */
    salign = 6;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            20, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            20, 21, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            20, 21, 22, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 15, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s6 d12: ================================================= */
    salign = 6;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d0: ================================================= */
    salign = 7;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 21, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 21, 22, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d1: ================================================= */
    salign = 7;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 21, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 21, 22, 
            23, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d2: ================================================= */
    salign = 7;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d3: ================================================= */
    salign = 7;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d4: ================================================= */
    salign = 7;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d5: ================================================= */
    salign = 7;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d6: ================================================= */
    salign = 7;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 16, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d7: ================================================= */
    salign = 7;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d8: ================================================= */
    salign = 7;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            16, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d9: ================================================= */
    salign = 7;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d10: ================================================= */
    salign = 7;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 21, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 21, 
            22, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 21, 
            22, 23, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 16, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d11: ================================================= */
    salign = 7;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            21, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            21, 22, 23, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s7 d12: ================================================= */
    salign = 7;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 21, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 21, 22, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 21, 22, 23, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 16, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d0: ================================================= */
    salign = 8;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d1: ================================================= */
    salign = 8;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 22, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 22, 23, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 22, 23, 
            24, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d2: ================================================= */
    salign = 8;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d3: ================================================= */
    salign = 8;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d4: ================================================= */
    salign = 8;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d5: ================================================= */
    salign = 8;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 17, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d6: ================================================= */
    salign = 8;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d7: ================================================= */
    salign = 8;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 17, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d8: ================================================= */
    salign = 8;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d9: ================================================= */
    salign = 8;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 17, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d10: ================================================= */
    salign = 8;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d11: ================================================= */
    salign = 8;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            22, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            22, 23, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            22, 23, 24, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 17, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s8 d12: ================================================= */
    salign = 8;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d0: ================================================= */
    salign = 9;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 23, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 23, 24, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d1: ================================================= */
    salign = 9;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 23, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 23, 24, 
            25, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d2: ================================================= */
    salign = 9;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 23, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 23, 
            24, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 23, 
            24, 25, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d3: ================================================= */
    salign = 9;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            23, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            23, 24, 25, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d4: ================================================= */
    salign = 9;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d5: ================================================= */
    salign = 9;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d6: ================================================= */
    salign = 9;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 18, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d7: ================================================= */
    salign = 9;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d8: ================================================= */
    salign = 9;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            18, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d9: ================================================= */
    salign = 9;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d10: ================================================= */
    salign = 9;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 23, 
            24, 25, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 18, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d11: ================================================= */
    salign = 9;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            23, 24, 25, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s9 d12: ================================================= */
    salign = 9;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 23, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 23, 24, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 23, 24, 25, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 18, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d0: ================================================ */
    salign = 10;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d1: ================================================ */
    salign = 10;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 24, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 24, 25, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 24, 25, 
            26, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d2: ================================================ */
    salign = 10;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d3: ================================================ */
    salign = 10;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            24, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            24, 25, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            24, 25, 26, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d4: ================================================ */
    salign = 10;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d5: ================================================ */
    salign = 10;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 24, 25, 26, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 19, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            30, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d6: ================================================ */
    salign = 10;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d7: ================================================ */
    salign = 10;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 24, 25, 26, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 19, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d8: ================================================ */
    salign = 10;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d9: ================================================ */
    salign = 10;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 24, 25, 
            26, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 19, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d10: ================================================ */
    salign = 10;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d11: ================================================ */
    salign = 10;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            24, 25, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            24, 25, 26, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 19, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s10 d12: ================================================ */
    salign = 10;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d0: ================================================ */
    salign = 11;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 25, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 25, 26, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 31, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d1: ================================================ */
    salign = 11;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 25, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 25, 26, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 25, 26, 
            27, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d2: ================================================ */
    salign = 11;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 25, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 25, 
            26, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 25, 
            26, 27, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 31, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d3: ================================================ */
    salign = 11;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            25, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            25, 26, 27, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d4: ================================================ */
    salign = 11;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 25, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 25, 26, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 25, 26, 27, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 31, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d5: ================================================ */
    salign = 11;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 25, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 25, 26, 27, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            31, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d6: ================================================ */
    salign = 11;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 25, 26, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 25, 26, 27, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            30, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 20, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            30, 31, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d7: ================================================ */
    salign = 11;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 25, 26, 27, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d8: ================================================ */
    salign = 11;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 25, 26, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            20, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 31, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d9: ================================================ */
    salign = 11;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 25, 26, 
            27, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d10: ================================================ */
    salign = 11;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 25, 
            26, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 25, 
            26, 27, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 20, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 31, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d11: ================================================ */
    salign = 11;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s11 d12: ================================================ */
    salign = 11;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 25, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 25, 26, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 25, 26, 27, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 20, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 31, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d0: ================================================ */
    salign = 12;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 32, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d1: ================================================ */
    salign = 12;
    dalign = 1;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 26, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 26, 27, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 26, 27, 
            28, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 31, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 31, 32, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d2: ================================================ */
    salign = 12;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 32, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d3: ================================================ */
    salign = 12;
    dalign = 3;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            26, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            26, 27, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            26, 27, 28, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 31, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 31, 32, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d4: ================================================ */
    salign = 12;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d5: ================================================ */
    salign = 12;
    dalign = 5;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 26, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 26, 27, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 26, 27, 28, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 31, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 21, 22, 23, 
            24, 25, 26, 27, 28, 29, 30, 31, 
            32, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d6: ================================================ */
    salign = 12;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            31, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            31, 32, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d7: ================================================ */
    salign = 12;
    dalign = 7;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 26, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 26, 27, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 26, 27, 28, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            30, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            30, 31, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 21, 
            22, 23, 24, 25, 26, 27, 28, 29, 
            30, 31, 32, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d8: ================================================ */
    salign = 12;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 32, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d9: ================================================ */
    salign = 12;
    dalign = 9;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 26, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 26, 27, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 26, 27, 
            28, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 31, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 21, 22, 23, 24, 25, 26, 27, 
            28, 29, 30, 31, 32, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d10: ================================================ */
    salign = 12;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 32, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d11: ================================================ */
    salign = 12;
    dalign = 11;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            26, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            26, 27, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            26, 27, 28, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 31, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 21, 22, 23, 24, 25, 
            26, 27, 28, 29, 30, 31, 32, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }

    /* Byte: Align s12 d12: ================================================ */
    salign = 12;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 1, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 3, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 5, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 7, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 9, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 11, expected, ARRAY_SIZE(expected),
                         isByte);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOByteMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isByte);
    }
}

/* ///////////////////////////////////////////////////////////////////////// */

static void
testOverlappingCharShort(CVMBool isChar)
{
    int salign, dalign;

    /* Char: Align s0 d0: ============================================== */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s0 d2: ============================================== */
    salign = 0;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 9, 10, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 9, 10, 11, 12, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 9, 10, 11, 12, 13, 14, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 9, 10, 11, 12, 13, 14, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 9, 10, 11, 12, 13, 14, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 9, 10, 11, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s0 d4: ============================================== */
    salign = 0;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            13, 14, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            13, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            13, 14, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s0 d6: ============================================== */
    salign = 0;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 9, 10, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 9, 10, 
            11, 12, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 9, 10, 
            11, 12, 13, 14, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 9, 10, 
            11, 12, 13, 14, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 9, 10, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 9, 10, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s0 d8: ============================================== */
    salign = 0;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s0 d10: ============================================== */
    salign = 0;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 9, 10, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 9, 10, 11, 12, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 9, 10, 11, 12, 13, 14, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 9, 10, 11, 12, 13, 14, 
            15, 16, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 9, 10, 11, 12, 13, 14, 
            15, 16, 17, 18, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 9, 10, 11, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s0 d12: ============================================== */
    salign = 0;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            13, 14, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            13, 14, 15, 16, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            13, 14, 15, 16, 17, 18, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s2 d0: ============================================== */
    salign = 2;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            11, 12, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            11, 12, 13, 14, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            11, 12, 13, 14, 15, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s2 d2: ============================================== */
    salign = 2;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s2 d4: ============================================== */
    salign = 2;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 11, 12, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 11, 12, 13, 14, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 11, 12, 13, 14, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 11, 12, 13, 14, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 11, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 11, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s2 d6: ============================================== */
    salign = 2;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 11, 12, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 11, 12, 
            13, 14, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 11, 12, 
            13, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 11, 12, 
            13, 14, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 11, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 11, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s2 d8: ============================================== */
    salign = 2;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            11, 12, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            11, 12, 13, 14, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            11, 12, 13, 14, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            11, 12, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s2 d10: ============================================== */
    salign = 2;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 11, 12, 13, 14, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 11, 12, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 11, 12, 13, 14, 15, 16, 
            17, 18, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s2 d12: ============================================== */
    salign = 2;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 11, 12, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 11, 12, 13, 14, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 11, 12, 13, 14, 
            15, 16, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 11, 12, 13, 14, 
            15, 16, 17, 18, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 11, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 11, 12, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s4 d0: ============================================== */
    salign = 4;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s4 d2: ============================================== */
    salign = 4;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s4 d4: ============================================== */
    salign = 4;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s4 d6: ============================================== */
    salign = 4;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 13, 14, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s4 d8: ============================================== */
    salign = 4;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s4 d10: ============================================== */
    salign = 4;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 13, 14, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s4 d12: ============================================== */
    salign = 4;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s6 d0: ============================================== */
    salign = 6;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s6 d2: ============================================== */
    salign = 6;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s6 d4: ============================================== */
    salign = 6;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s6 d6: ============================================== */
    salign = 6;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s6 d8: ============================================== */
    salign = 6;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            15, 16, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s6 d10: ============================================== */
    salign = 6;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s6 d12: ============================================== */
    salign = 6;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 15, 16, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s8 d0: ============================================== */
    salign = 8;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s8 d2: ============================================== */
    salign = 8;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s8 d4: ============================================== */
    salign = 8;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s8 d6: ============================================== */
    salign = 8;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 17, 18, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s8 d8: ============================================== */
    salign = 8;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s8 d10: ============================================== */
    salign = 8;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 17, 18, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s8 d12: ============================================== */
    salign = 8;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s10 d0: ============================================== */
    salign = 10;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s10 d2: ============================================== */
    salign = 10;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s10 d4: ============================================== */
    salign = 10;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s10 d6: ============================================== */
    salign = 10;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s10 d8: ============================================== */
    salign = 10;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            19, 20, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s10 d10: ============================================== */
    salign = 10;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s10 d12: ============================================== */
    salign = 10;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 19, 20, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s12 d0: ============================================== */
    salign = 12;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 32, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s12 d2: ============================================== */
    salign = 12;
    dalign = 2;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 32, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s12 d4: ============================================== */
    salign = 12;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s12 d6: ============================================== */
    salign = 12;
    dalign = 6;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 21, 22, 
            23, 24, 25, 26, 27, 28, 29, 30, 
            31, 32, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s12 d8: ============================================== */
    salign = 12;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 32, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s12 d10: ============================================== */
    salign = 12;
    dalign = 10;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 21, 22, 23, 24, 25, 26, 
            27, 28, 29, 30, 31, 32, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }

    /* Char: Align s12 d12: ============================================== */
    salign = 12;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 2, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 6, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 10, expected, ARRAY_SIZE(expected),
                         isChar);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOCharMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isChar);
    }
}

/* ///////////////////////////////////////////////////////////////////////// */

static void
testOverlappingIntFloat(CVMBool isInt)
{
    int salign, dalign;

    /* Int: Align s0 d0: ============================================== */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s0 d4: ============================================== */
    salign = 0;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            13, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 9, 10, 11, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s0 d8: ============================================== */
    salign = 0;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s0 d12: ============================================== */
    salign = 0;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            13, 14, 15, 16, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 9, 10, 11, 12, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s4 d0: ============================================== */
    salign = 4;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s4 d4: ============================================== */
    salign = 4;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s4 d8: ============================================== */
    salign = 4;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            13, 14, 15, 16, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s4 d12: ============================================== */
    salign = 4;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s8 d0: ============================================== */
    salign = 8;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s8 d4: ============================================== */
    salign = 8;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s8 d8: ============================================== */
    salign = 8;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s8 d12: ============================================== */
    salign = 8;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 17, 18, 19, 20, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s12 d0: ============================================== */
    salign = 12;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 32, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s12 d4: ============================================== */
    salign = 12;
    dalign = 4;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s12 d8: ============================================== */
    salign = 12;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            21, 22, 23, 24, 25, 26, 27, 28, 
            29, 30, 31, 32, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

    /* Int: Align s12 d12: ============================================== */
    salign = 12;
    dalign = 12;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 4, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isInt);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
        };
        testOIntMemmove(salign, dalign, 12, expected, ARRAY_SIZE(expected),
                         isInt);
    }

}

/* ///////////////////////////////////////////////////////////////////////// */

static void
testOverlappingLongDouble(CVMBool isLong)
{
    int salign, dalign;

    /* Long: Align s0 d0: ============================================== */
    salign = 0;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s0 d8: ============================================== */
    salign = 0;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s0 d16: ============================================== */
    salign = 0;
    dalign = 16;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s8 d0: ============================================== */
    salign = 8;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s8 d8: ============================================== */
    salign = 8;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s8 d16: ============================================== */
    salign = 8;
    dalign = 16;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s16 d0: ============================================== */
    salign = 16;
    dalign = 0;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s16 d8: ============================================== */
    salign = 16;
    dalign = 8;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

    /* Long: Align s16 d16: ============================================== */
    salign = 16;
    dalign = 16;
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 0, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 8, expected, ARRAY_SIZE(expected),
                         isLong);
    }
    {
        static const CVMUint8 expected[] = {
            1, 2, 3, 4, 5, 6, 7, 8, 
            9, 10, 11, 12, 13, 14, 15, 16, 
            17, 18, 19, 20, 21, 22, 23, 24, 
            25, 26, 27, 28, 29, 30, 31, 32, 
            33, 34, 35, 36, 37, 38, 39, 40, 
            41, 42, 43, 44, 45, 46, 47, 48, 
        };
        testOLongMemmove(salign, dalign, 16, expected, ARRAY_SIZE(expected),
                         isLong);
    }

}

void testMemmove(void)
{
    testCount = 0;
    testPassed = 0;

    testByteBoolean(CVM_TRUE);
    testByteBoolean(CVM_FALSE);
    testCharShort(CVM_TRUE);
    testCharShort(CVM_FALSE);
    testIntFloat(CVM_TRUE);
    testIntFloat(CVM_FALSE);
    testLongDouble(CVM_TRUE);
    testLongDouble(CVM_FALSE);

    testOverlappingByteBoolean(CVM_TRUE);
    testOverlappingByteBoolean(CVM_FALSE);
    testOverlappingCharShort(CVM_TRUE);
    testOverlappingCharShort(CVM_FALSE);
    testOverlappingIntFloat(CVM_TRUE);
    testOverlappingIntFloat(CVM_FALSE);
    testOverlappingLongDouble(CVM_TRUE);
    testOverlappingLongDouble(CVM_FALSE);

    /* Dump Results: */
    CVMconsolePrintf("\n%d passed\n%d failed\n%d total\n",
                     testPassed, testCount - testPassed, testCount);

}

#undef ARRAY_SIZE
